#!/usr/bin/env python
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY  ; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

# Pytigon - wxpython and django application framework

# author: "Slawomir Cholaj (slawomir.cholaj@gmail.com)"
# license: "LGPL 3.0"


"""This is the main Pytigon client moudule. Function :func:`~pytigon_gui.pytigon.main` create SchApp object, witch extends wxPython wx.App.
Function :func:`~pytigon_gui.pytigon.main` process pytigon command line arguments. Module supports:

- instalation of pytigon applications,

- Remote Procedure Calling protocol

- login to server process
"""

import os

# os.environ["G_DEBUG"] = "fatal-criticals"
import sys
import time
import platform
import zipfile
import getopt
from multiprocessing import Process
import configparser
from urllib.parse import urljoin
import pytigon

# import debugpy
# debugpy.listen(("localhost", 5678))
# debugpy.wait_for_client()

if platform.system() == "Windows":
    # grouping pytigon applicactions in the windows taskbar
    import ctypes

    myappid = "slawomir_cholaj.pytigon.main.01"
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

CWD_PATH = os.path.join(os.getcwd(), "..")
SRC_PATH = os.path.dirname(pytigon.__file__)

ROOT_PATH = SRC_PATH

if ROOT_PATH.startswith("."):
    ROOT_PATH = CWD_PATH + "/" + ROOT_PATH
sys.path.append(ROOT_PATH)
sys.path.append(ROOT_PATH + "/appdata")

os.environ["EMBEDED_DJANGO_SERVER"] = "1"

from pytigon_lib import init_paths

init_paths()

if platform.system() == "Windows":
    sys.path.insert(0, ROOT_PATH + "/ext_lib_cli_win")
else:
    sys.path.insert(0, ROOT_PATH + "/ext_lib_cli_lin")

_INSPECTION = False
_DEBUG = False
_TRACE = False
_VIDEO = False
_VIDEO_NAME = ""
_APP_SIZE = (1024, 768)
_RPC = False
_WEBSOCKET = None


def usage():
    print(process_argv.__doc__)


def process_argv(argv):
    """Run pytigon application: pytigon [option]... arg

    command line arguments:
        arg: project name, schdevtools for example
             or address of http server, http://www.pytigon.cloud for example
             or pytigon script name, test.schdevtools.schsimplescripts.ptig for example
             or pytigon installation file name
             or django management command in format: manage_[[package name]], manage_schdevtools runserver for example
             or python script, test.py for example

        options:
            -h, --help - show help

            -b --embededbrowser - run embeded browseer window
            -s --embededserver - run in background embeded http server
            --listen=<address>:<port> - set address and port for embeded serwer
            --menu_always - show menu event then configuration says otherwise
            --no_splash - do not show splash window
            --no_gui - run program without gui
            --server_only - run only http server
            --channels - start with channels library
            --websocket_id=relative address - embed websocket client, address can be mulitipart,
                           parts separated by a semicolon
            --no_gui - run program without gui

            --migrate - run django command: manage.py migrate
            --loaddb - run django command: manage.py loaddb

            --rpc=<port> - set tcp port of rpc
            --video - record video session
            -p <parameters>, --param=<parameters> - parametres of request to http server

            -d --debug - debug mode

            --inspection - turn on wxPython inspection
            --trace - show trace of python calls
    """
    try:
        (opts, args) = getopt.gnu_getopt(
            argv,
            "h:dmpbsu:p:",
            [
                "help",
                "username=",
                "password=",
                "embededbrowser",
                "embededserver",
                "embededtaskqueue",
                "websocket_id=",
                "channels",
                "migrate",
                "loaddb",
                "server_only",
                "debug",
                "rpc=",
                "param=",
                "inspection",
                "trace",
                "video=",
                "no_gui",
                "no_splash",
                "menu_always",
                "listen=",
                "import=",
                "extra=",
            ],
        )
    except getopt.GetoptError:
        usage()
        return None

    ret = {"args": args}

    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
            return None
        elif opt == "--migrate":
            ret["sync"] = True
        elif opt == "--loaddb":
            ret["loaddb"] = True
        elif opt == "--server_only":
            ret["server_only"] = True
        elif opt in ("-u--username"):
            ret["username"] = arg
        elif opt in ("-p--password"):
            ret["password"] = arg
        elif opt in ("--listen",):
            ret["listen"] = arg
        elif opt in ("--extra",):
            ret["extra"] = arg
        elif opt in ("--import",):
            ret["import"] = arg
        elif opt in ("-b--embededbrowser"):
            ret["embeded_browser"] = True
        elif opt in ("-s", "--embededserver"):
            ret["address"] = "embeded"
            ret["extern_prj"] = True
        elif opt in ("--embededtaskqueue",):
            ret["embeded_taskqueue"] = True
        elif opt == "--no_gui":
            ret["nogui"] = True
        elif opt in ("--menu_always",):
            ret["menu_always"] = True
        elif opt in ("--rpc"):
            ret["rpc"] = int(arg)
        elif opt in ("-p", "--param"):
            ret["param"] = arg
        elif opt in ("--channels"):
            ret["channels"] = True
        elif opt in ("--websocket_id",):
            ret["websocket"] = arg
        elif opt in ("--no_splash"):
            ret["no_splash"] = True
        elif opt in ("-d", "--debug"):
            global _DEBUG
            _DEBUG = True
        elif opt in ("--inspection",):
            global _INSPECTION
            _INSPECTION = True
        elif opt in ("--trace",):
            global _TRACE
            _TRACE = True
        elif opt in ("--video",):
            global _VIDEO, _APP_SIZE, _VIDEO_NAME
            _APP_SIZE = (1280, 720)
            _VIDEO = True
            _VIDEO_NAME = arg
            if not _VIDEO_NAME:
                _VIDEO_NAME = "video.avi"
    return ret


_PARAM = process_argv(sys.argv[1:])
if _PARAM == None:
    sys.exit(0)

import asyncio
from asyncio import get_event_loop, set_event_loop_policy

if sys.platform == "win32":
    set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

LOOP = get_event_loop()

from pytigon_lib.schtools.main_paths import get_main_paths

if "PRJ_NAME" in os.environ:
    PRJ_NAME = os.environ["PRJ_NAME"]
    PATHS = get_main_paths(os.environ["PRJ_NAME"])
else:
    PRJ_NAME = None
    PATHS = get_main_paths()

sys.path.append(PATHS["PRJ_PATH"])

from pytigon_lib.schtools.install_init import init

DATA_PATH = PATHS["DATA_PATH"]
PYTIGON_PATH = PATHS["PYTIGON_PATH"]

init(
    "_schall",
    PATHS["ROOT_PATH"],
    PATHS["DATA_PATH"],
    PATHS["PRJ_PATH"],
    PATHS["STATIC_PATH"],
    [PATHS["MEDIA_PATH"], PATHS["UPLOAD_PATH"]] if PRJ_NAME else None,
)

import wx

_ = wx.GetTranslation


def process_adv_argv():
    global _PARAM
    if not ("args" in _PARAM and len(_PARAM["args"]) > 0):
        _app = wx.App()

        choices = [
            ff
            for ff in os.listdir(PATHS["PRJ_PATH"])
            if not ff.startswith("_")
            and os.path.isdir(os.path.join(PATHS["PRJ_PATH"], ff))
            and os.path.exists(os.path.join(PATHS["PRJ_PATH"], ff, "settings_app.py"))
        ]
        dlg = wx.SingleChoiceDialog(
            None,
            _("select the application to run"),
            _("Pytigon"),
            choices,
            wx.CHOICEDLG_STYLE,
        )
        if dlg.ShowModal() == wx.ID_OK:
            arg = dlg.GetStringSelection()
            dlg.Destroy()

        else:
            dlg.Destroy()
            sys.exit(0)

        _app.MainLoop()
        _app = None

    else:
        arg = _PARAM["args"][0].strip()
    if not (arg == "embeded" or "." in arg or "/" in arg):
        CWD_PATH = os.path.join(PATHS["PRJ_PATH"], arg)
        if not os.path.exists(os.path.join(CWD_PATH, "settings_app.py")):
            print(_("Application pack: '%s' does not exists") % arg)
            sys.exit(0)
        else:
            sys.path.insert(0, CWD_PATH)

            import settings_app

            if hasattr(settings_app, "GUI_COMMAND_LINE"):
                x = settings_app.GUI_COMMAND_LINE.split(" ")
                param = process_argv(x)
                for key, value in param.items():
                    if not key in _PARAM:
                        _PARAM[key] = value


process_adv_argv()

if "channels" in _PARAM or "rpc" in _PARAM or "websocket" in _PARAM:
    try:
        from wxasync import AsyncBind, WxAsyncApp, StartCoroutine
    except:
        asyncio.futures.CancelledError = asyncio.CancelledError
        from wxasync import AsyncBind, WxAsyncApp, StartCoroutine

    class SChAsyncApp(WxAsyncApp):
        async def MainLoop(self):
            evtloop = wx.GUIEventLoop()
            with wx.EventLoopActivator(evtloop):
                while not self.exiting:
                    if platform.system() == "Darwin":
                        evtloop.DispatchTimeout(0)
                    else:
                        while self.HasPendingEvents():
                            self.ProcessPendingEvents()
                        while evtloop.Pending():
                            evtloop.Dispatch()
                    await asyncio.sleep(0.005)
                    evtloop.ProcessIdle()


from pytigon_gui.guilib import image
from pytigon_gui.guilib import pytigon_install
from pytigon_gui.guilib.logindialog import LoginDialog
from pytigon_gui.guiframe import appframe
from pytigon_gui.guilib.threadwindow import SchThreadManager
from pytigon_lib.schfs.vfstools import extractall
from pytigon_lib.schparser.html_parsers import SimpleTabParser
from pytigon_gui.guilib.tools import standard_tab_colour, colour_to_html
from pytigon_lib.schhttptools import httpclient
from pytigon_gui.guilib.httperror import http_error
from pytigon_lib.schhttptools.httpclient import HttpResponse
from pytigon_gui.guiframe import browserframe
from pytigon_lib.schtools import createparm
from pytigon_lib.schparser.html_parsers import ShtmlParser
from pytigon_lib.schtools.schjson import json_dumps, json_loads
import pytigon_gui.guictrl.tag

if "rpc" in _PARAM or "websocket" in _PARAM:
    import twisted.internet.asyncioreactor

    twisted.internet.asyncioreactor.install()

    from twisted.internet import reactor
    import twisted

    if "rpc" in _PARAM:
        from twisted.web import xmlrpc, server

        _RPC = True
    if "websocket" in _PARAM:
        # from autobahn.twisted.websocket import WebSocketClientFactory, WebSocketClientProtocol, connectWS
        from pytigon_lib.schhttptools.websocket import create_websocket_client

        _WEBSOCKET = _PARAM["websocket"]

if not "channels" in _PARAM:
    os.environ["PYTIGON_WITHOUT_CHANNELS"] = "1"

# import gc
# gc.set_debug(gc.DEBUG_STATS | gc.DEBUG_LEAK)
# gc.disable()

wx.RegisterId(10000)
wx.outputWindowClass = None

if _INSPECTION:
    os.environ["GTK_THEME"] = "Adwaita"

    import wx.lib.mixins.inspection

    if "channels" in _PARAM or "rpc" in _PARAM or "websocket" in _PARAM:

        class InspectableApp(SChAsyncApp, wx.lib.mixins.inspection.InspectionMixin):
            def OnInit(self):
                self.InitInspection()
                return True

        App = InspectableApp
    else:
        App = wx.lib.mixins.inspection.InspectableApp

    if _TRACE:

        def trace_calls(frame, event, arg):
            if event != "call":
                return
            co = frame.f_code
            func_name = co.co_name
            if func_name == "write":
                # Ignore write() calls from print statements
                return
            for pos in ("process_window_event", "idle", "timer", "update_ui"):
                if pos in func_name:
                    return
            if (
                "process_window_event" in func_name
                or "idle" in func_name
                or "idle" in func_name
            ):
                return
            func_line_no = frame.f_lineno
            func_filename = co.co_filename
            if not "wx/core" in func_filename:
                return
            caller = frame.f_back
            caller_line_no = caller.f_lineno
            caller_filename = caller.f_code.co_filename
            print(
                "Call to %s on line %s of %s from line %s of %s"
                % (
                    func_name,
                    func_line_no,
                    func_filename,
                    caller_line_no,
                    caller_filename,
                )
            )
            return

        sys.settrace(trace_calls)

else:
    if "channels" in _PARAM or "rpc" in _PARAM or "websocket" in _PARAM:
        App = SChAsyncApp
    else:
        App = wx.App

if _RPC:
    _BASE_APP = xmlrpc.XMLRPC
else:

    class _base:
        pass

    _BASE_APP = _base


class SchApp(App, _BASE_APP):
    """It is a subclass of wxPython wx.App. Depending on the command line parameters it is also subclass of
    xmlrpc.XMLRPC and wx.lib.mixins.inspection.InspectableApp.
    """

    def __init__(self):
        """Construct an application."""
        global _PARAM
        App.__init__(self)
        if _RPC:
            xmlrpc.XMLRPC.__init__(self)

        if (
            not "no_splash" in _PARAM
            and not "nogui" in _PARAM
            and not "server_only" in _PARAM
        ):
            # bitmap = wx.Bitmap(SRC_PATH + "/pytigon.svg", wx.BITMAP_TYPE_JPEG)
            img = wx.svg.SVGimage.CreateFromFile(SRC_PATH + "/pytigon.svg")
            # img.ConvertAlphaToMask()
            bitmap = img.ConvertToBitmap(
                scale=2, width=int(img.width * 2), height=int(img.height * 2)
            )
            wx.BITMAP_TYPE_PNG
            splash = wx.adv.SplashScreen(
                bitmap,
                wx.adv.SPLASH_CENTRE_ON_SCREEN | wx.adv.SPLASH_TIMEOUT,
                1000,
                None,
                -1,
                wx.DefaultPosition,
                wx.DefaultSize,
                wx.BORDER_SIMPLE | wx.STAY_ON_TOP,
            )
            splash.Update()
            wx.Yield()

        config_name = os.path.join(SRC_PATH, "pytigon.ini")
        self.config = configparser.ConfigParser()
        self.config.read(config_name)

        self.app_size = _APP_SIZE

        self.locale = None
        self.ext_app = []
        self.ext_app_http = {}
        self.lock = None

        self.base_address = None

        self.src_path = SRC_PATH
        self.root_path = ROOT_PATH
        self.cwd_path = CWD_PATH
        self.data_path = DATA_PATH
        self.pytigon_path = PYTIGON_PATH

        self.http = None
        self.images = None
        self.mp = None

        self.server = None
        self.cwd = None
        self.inst_dir = None
        self.start_pages = []

        self.csrf_token = None
        self.title = None
        self.plugins = None
        self.extern_data = {}

        self.thread_manager = None
        self.task_manager = None

        self.menu_always = False
        self.authorized = False
        self.rpc = None

        self.websockets = {}
        self.websockets_callbacks = {}

        self.gui_style = (
            "app.gui_style = tree(toolbar(file(exit,open),clipboard, statusbar))"
        )

        self.COLOUR_HIGHLIGHT = colour_to_html(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        )
        self.COLOUR_BACKGROUND = colour_to_html(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE)
        )
        self.COLOUR_SHADOW = colour_to_html(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)
        )
        self.COLOUR_DKSHADOW = colour_to_html(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DDKSHADOW)
        )
        self.COLOUR_ACTIVECATPION = colour_to_html(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_ACTIVECAPTION)
        )
        self.COLOUR_INFOBK = colour_to_html(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_INFOBK)
        )

        self.ctrl_process = {}

        class callback:
            def on_websocket_message(msg):
                # pass
                print("CALLBACK:", msg)

        self.add_websoket_callback("/schbuilder/clock/channel/", callback)

    def register_ctrl_process_fun(self, tag, fun):
        """Register function, which is called when widget connected to the specified tag is created.

        Args:
            tag: it must be the name of class in pytigon_gui/guictrl/schctrl typed in lowercase, preceded by "ctrl".
            For example: ctrlgrid, ctrltable, ctrllistbox are valid tag names.

            fun: a callback function with is called with one parameter - widget defined in pytigon_gui/guictrl/schctrl instance.

        Returns:
            None

        Example:
            see: appdata/plugins/standard/keymap/__init__.py
        """

        if tag in self.ctrl_process:
            self.ctrl_process[tag].append(fun)
        else:
            self.ctrl_process[tag] = [fun]

    # some XML-RPC function calls for twisted server
    def xmlrpc_stop(self):
        """Closes the wx application."""
        self.frame.Close()  # Sending closing event
        return "Shutdown initiated"

    def xmlrpc_title(self, title):
        """Set the main window title

        Args:
            title - new title of application top window
        """
        self.GetTopWindow().SetTitle(title)
        return title

    def get_locale_object(self):
        """Get locale object

        Returns:
            application wx.Locale object
        """
        return self.locale

    def _init2(self, address, app):
        self.base_address = address
        self.base_app = app
        if self.base_app:
            self.base_path = urljoin(self.base_address, self.base_app)
        else:
            self.base_path = self.base_address
        if app and app != "":
            href = "/" + app + "/"
        else:
            href = "/"

        self.http = httpclient.AppHttp(address + "/", self)
        response = self.http.get(self, href)
        if response.ret_code != 200:
            if not login(href, auth_type="basic"):
                return 0
            else:
                self.authorized = True
            response = self.http.get(href, "/")

        if response.ret_code != 200:
            return response

        if app and app != "":
            self.images = image.SchImage("/" + app + "/site_media/app.png")
            response = self.http.get(self, "/" + app + "/")
        else:
            self.images = image.SchImage("/site_media/app.png")
            response = self.http.get(self, "/")

        ret_str = response.str()

        self.mp = SimpleTabParser()
        self.mp.feed(ret_str)
        self.mp.close()

        return response.ret_code

    def _re_init(self, address, app):
        self.base_address = address
        self.base_app = app
        if self.base_app:
            self.base_path = urljoin(self.base_address, self.base_app)
        else:
            self.base_path = self.base_address

        if app and app != "":
            response = self.http.get(self, "/" + app + "/")
        else:
            response = self.http.get(self, "/")
        ret = response.str()
        self.mp = SimpleTabParser()
        self.mp.feed(ret)
        self.mp.close()

    async def test_websockets(self):
        print("-----------------------------------------------------------------")
        print(self.websockets)
        await self.websocket_send("/schtasks/show_task_events/channel/", {"id": "test"})
        print("=================================================================")

        count = 999
        while True:
            await asyncio.sleep(1)
            await self.websocket_send(
                "/schbuilder/clock/channel/", {"title": "Hello world %s" % count}
            )
            count -= 1

    async def init_websockets(self):
        from channels.generic.websocket import WebsocketConsumer, AsyncWebsocketConsumer

        obj = self
        old_init = WebsocketConsumer.__init__
        old_init_async = AsyncWebsocketConsumer.__init__

        def accept(self):
            pass

        async def accept_async(self):
            pass

        def init(self, *args, **kwargs):
            old_init(self, *args, **kwargs)
            self.connect()

        def init_async(self, *args, **kwargs):
            nonlocal obj
            old_init_async(self, *args, **kwargs)
            if hasattr(self, "Bind"):
                obj.StartCoroutine(self.connect, self)
            else:

                def _bind(x, y, z):
                    pass

                self.Bind = _bind

        WebsocketConsumer.__init__ = init
        AsyncWebsocketConsumer.__init__ = init_async
        WebsocketConsumer.accept = accept
        AsyncWebsocketConsumer.accept = accept_async

        tasks = []
        if self.websockets:
            for key, value in self.websockets.items():
                if value.status == 1:
                    value.status = 2
                    tasks.append(
                        asyncio.create_task(
                            httpclient.local_websocket(
                                self.base_address.replace("http://", "ws://") + key,
                                value.input_queue,
                                value,
                            )
                        )
                    )
            if tasks:
                # try:
                done, pending = await asyncio.wait(tasks)
                # done, pending = yield from asyncio.wait([raise_exception()], timeout=1)
                assert not pending
                (future,) = done  # unpack a set of length one
                print(future.result())  # raise an exception or use future.exception()
            # except Error:
            # except:
            #    print('got exception', flush=True)
            # else:
            #    print('no exception', flush=True)

    def create_websocket(self, websocket_id, callback):
        local = True if app.base_address.startswith("http://127.0.0.2") else False
        create_websocket_client(self, websocket_id, local, callback)
        if local:
            tasks = []
            if self.websockets:
                for key, value in self.websockets.items():
                    if value.status == 1:
                        value.status = 2
                        tasks.append(
                            httpclient.local_websocket(
                                self.base_address.replace("http://", "ws://") + key,
                                value.input_queue,
                                value,
                            )
                        )
                if tasks:

                    async def reinit_websockets():
                        nonlocal tasks
                        done, pending = await asyncio.wait(tasks)
                        # done, pending = yield from asyncio.wait([raise_exception()], timeout=1)
                        assert not pending
                        (future,) = done  # unpack a set of length one
                        print(
                            future.result()
                        )  # raise an exception or use future.exception(

                    self.StartCoroutine(reinit_websockets, self.GetTopWindow())

        # if local:
        #    self.StartCoroutine(self.init_websockets, self)

    def make_href(self, href):
        if self.base_app and href.startswith("/"):
            return "/" + self.base_app + href
        else:
            return href

    def SetTopWindow(self, frame):
        wx.App.SetTopWindow(self, frame)
        icon = wx.Icon(SRC_PATH + "/pytigon.ico", wx.BITMAP_TYPE_ICO)
        frame.SetIcon(icon)

        if hasattr(frame, "statusbar") and frame.statusbar:
            self.thread_manager = SchThreadManager(self, frame.statusbar)
        if _INSPECTION:
            self.ShowInspectionTool()

        if hasattr(self, "StartCoroutine"):
            if self.base_address and self.base_address.startswith("http://127.0.0.2"):
                wx.CallAfter(self.StartCoroutine, self.init_websockets, frame)
            # if _DEBUG:
            #    self.StartCoroutine(self.test_websockets, frame)

    def register_extern_app(self, address, app):
        self.ext_app.append((app, address))

    def get_http_for_adr(self, address):
        for app in self.ext_app:
            if address.upper().startswith(app[1]):
                if app[0] in self.ext_app_http:
                    return self.ext_app_http[app[0]]
                else:
                    http = httpclient.AppHttp(app[1], app[0])
                    self.exft_app_http[app[0]] = http
                    return http
        return self.http

    def get_http(self, win):
        """Get :class:`~pytigon_lib.schhttptools.httpclient.HttpClient` object

        With returned object you can directly make requests to http server.

        Args:
            win - wx.Window derived object

        Returns:
            :class:`~pytigon_lib.schhttptools.httpclient.HttpClient` object
        """
        winparent = win
        while winparent:
            if hasattr(winparent, "get_app_http"):
                return winparent.get_app_http()
            winparent = winparent.GetParent()
        return self.http

    def read_html(self, win, address_or_parser, parameters):
        """Prepare request to http server and read result

        Args:
            win: wx.Window derived object
            address_or_parser: can be: address of read_htmlhttp page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'
            parameters: dict

        Returns :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser' object
        """
        if type(address_or_parser) == str:
            http = self.get_http(win)

            if parameters and type(parameters) == dict:
                adr = address_or_parser
                response = http.post(self, adr, parm=parameters)
            else:
                if parameters:
                    adrtmp = createparm.create_parm(address_or_parser, parameters)
                    if adrtmp:
                        adr = adrtmp[0] + adrtmp[1] + adrtmp[2]
                    else:
                        adr = address_or_parser
                else:
                    adr = address_or_parser
                response = http.get(win, adr)
            if response.ret_code == 404:
                raise Exception("http", "404")
            ptr = response.str()
            mp = ShtmlParser()
            mp.process(ptr, address_or_parser)
            mp.address = adr
        elif type(address_or_parser) == HttpResponse:
            adr = address_or_parser.url
            ptr = address_or_parser.str()
            mp = ShtmlParser()
            mp.process(ptr, adr)
            mp.address = adr
        else:
            adr = None
            if address_or_parser:
                mp = address_or_parser
            else:
                mp = ShtmlParser()
                mp.process("<html><body></body></html>")
                mp.address = None
        return (mp, adr)

    def get_tab(self, nr):
        """Return setup tab

        When pytigon client connect to server it receive html page with few configuration tables.
        This function return rendered table.

        Args:
            nr: table number

        Returns:
            table - :class:'~pytigon_lib.schparser.html_parsers.SimpleTabParser' object

        """
        if self.mp:
            return self.mp.tables[nr]
        else:
            return None

    # def get_active_window(self):
    #    return self.GetTopWindow()

    # def get_main_windows(self):
    #    return [self.GetTopWindow()]

    # def append_thread(self, thread_address):
    #    if self.thread_manager:
    #        self.thread_manager.append(thread_address)

    # def timer(self):
    #    if self.thread_manager:
    #        self.thread_manager.timer()

    def get_working_dir(self):
        """return pytigon working director - ~/pytigon_data"""
        return os.path.join(os.path.expanduser("~"), "pytigon_data/")

    def _get_parm_for_server(self):
        ret = ""
        for pos in standard_tab_colour():
            ret = ret + pos[0] + ":" + pos[1] + ","
        return ret[:-1].replace("#", "")

    def _install_plugins(self):
        home_dir = self.get_working_dir()
        p = self.plugins
        if p:
            sys.path.append(home_dir)
            for plugin in p:
                if not "/" in plugin:
                    continue
                if plugin == "" or plugin.startswith("standard/"):
                    continue
                print(plugin)
                app_name = plugin.split("/")[0]
                plugin_name = plugin.split("/")[1]
                if plugin_name == "install":
                    plugins_cache = ""
                else:
                    plugins_cache = "plugins_cache/"
                if not os.path.exists(home_dir + plugins_cache + str(app_name)):
                    os.mkdir(home_dir + plugins_cache + str(app_name))
                    ini = open(
                        home_dir + plugins_cache + str(app_name) + "/__init__.py", "w"
                    )
                    ini.write(" ")
                    ini.close()
                if not os.path.exists(home_dir + plugins_cache + str(plugin) + ".zip"):
                    http = wx.GetApp().http
                    response = http.get(self, "/schsys/plugins/" + str(plugin) + "/")
                    z_data = response.ptr()
                    x = open(home_dir + plugins_cache + str(plugin) + ".zip", "wb")
                    x.write(z_data)
                    x.close()
                    zip_name = home_dir + plugins_cache + str(plugin) + ".zip"
                    extract_to = home_dir + plugins_cache + str(app_name)
                    zip_handle = zipfile.ZipFile(zip_name)
                    extractall(zip_handle, extract_to)
                    zip_handle.close()

    def on_exit(self):
        if self.task_manager:
            self.task_manager.terminate()
            # if len(self.task_manager.list_threads(all=False)) > 0:
            #    dlg = wx.MessageDialog(
            #        None,
            #        _("There are background tasks - kill?"),
            #        _("Warning"),
            #        wx.YES_NO | wx.ICON_QUESTION,
            #    )
            #    result = dlg.ShowModal()
            #    if result == wx.ID_YES:
            #        self.task_manager.kill_all()

    def run_script(self, app_name, script_path):
        with open(script_path, "rb") as s:
            wx.CallAfter(
                self.GetTopWindow().new_main_page,
                "/" + app_name + "/run_script/",
                "Run script",
                {"script": s.read()},
            )

    def add_websoket_callback(self, websocket_id, callback):
        if websocket_id in self.websockets_callbacks:
            self.websockets_callbacks[websocket_id].append(callback)
        else:
            self.websockets_callbacks[websocket_id] = [callback]

    def remove_websocket_callback(self, websocket_id, callback):
        if websocket_id in self.websockets_callbacks:
            if callback in self.websockets_callbacks[websocket_id]:
                self.websockets_callbacks[websocket_id].remove(callback)

    async def websocket_send(self, websocket_id, msg):
        if websocket_id in self.websockets:
            if not "clock" in websocket_id:
                print(websocket_id)
            obj = self.websockets[websocket_id].send_message(msg)
            if obj:
                await obj

    def on_websocket_callback(self, client, event_name, argv):
        if client.websocket_id in self.websockets_callbacks:
            for callback in self.websockets_callbacks[client.websocket_id]:
                if hasattr(callback, event_name):
                    if "channel" in argv:
                        if hasattr(callback, "accept_channel"):
                            if not getattr(callback, "accept_channel")(argv["channel"]):
                                continue
                    getattr(callback, event_name)(**argv)

    def on_websocket_connect(self, client, websocket_id, response):
        return self.on_websocket_callback(
            client, "on_websocket_connect", {"response": response}
        )

    def on_websocket_open(self, client, websocket_id):
        return self.on_websocket_callback(client, "on_websocket_open", {})

    def on_websocket_message(self, client, websocket_id, msg):
        return self.on_websocket_callback(client, "on_websocket_message", msg)


def login(base_href, auth_type=None, username=None):
    """Show login form"""
    dlg = LoginDialog(None, 101, _("Pytigon - login"), username=username)

    while dlg.ShowModal() == wx.ID_OK:
        username = dlg.text1.GetValue()
        password = dlg.text2.GetValue()

        parm = {
            "username": username,
            "password": password,
            "next": "/schsys/ok/",
            "client_param": wx.GetApp()._get_parm_for_server(),
        }
        if auth_type == None:
            ret = wx.GetApp().http.post(
                wx.GetApp(),
                "/schsys/do_login/?from_pytigon=1",
                parm,
                credentials=(username, password),
            )

            ret_str = ret.str()

            if "$$RETURN_OK" in ret_str:
                dlg.Destroy()
                return True
            else:
                if not "id_password" in ret_str:
                    dlg.Destroy()
                    return False
                else:
                    dlg.message.SetLabel(_("Failed login attempt!"))
        else:
            result = wx.GetApp().http.get(
                wx.GetApp(), base_href, credentials=(username, password)
            )
            if result.ret_code == 200:
                dlg.Destroy()
                return True
            else:
                dlg.message.SetLabel(
                    _("Failed login attempt! http error: %s") % result.ret_code
                )
    dlg.Destroy()
    return False


def _main_init():
    global CWD_PATH, _PARAM, app

    args = _PARAM["args"]
    apps = []
    address = "http://127.0.0.2"
    app_title = _("Pytigon")
    app_name = ""

    app = SchApp()
    if "menu_always" in _PARAM:
        app.menu_always = True
    if "rpc" in _PARAM:
        app.rpc = int(_PARAM["rpc"])
    if "param" in _PARAM:
        app.param = _PARAM["param"]
    if "embeded_browser" in _PARAM:
        app.embeded_browser = True
    else:
        app.embeded_browser = False
    if "extern_prj" in _PARAM:
        extern_prj = True
    else:
        extern_prj = False
    if "address" in _PARAM:
        address = _PARAM["address"]

    os.environ["DJANGO_SETTINGS_MODULE"] = "settings_app"
    if len(args) > 0:
        if ".ptig" in args[0].lower():
            prg_name = args[0].replace("\\", "/").split("/")[-1]
            x = prg_name.split(".")
            if len(x) == 2 or (len(x) > 2 and x[-2].lower() == "inst"):
                prg_name2 = x[0]
                path = os.path.join(PATHS["PRJ_PATH_ALT"], "_schremote")
                sys.path.append(path)
                if not pytigon_install.install(args[0]):
                    return (None, None)
                # sys.path.remove(path)
                return (None, None)
                CWD_PATH = os.path.join(PATHS["PRJ_PATH"], prg_name2)
            else:
                if len(x) > 3:
                    prg_name2 = x[0]
                    app_name2 = x[-2]
                    prj = x[-3]
                    CWD_PATH = os.path.join(PATHS["PRJ_PATH"], prj.strip())
                    if not os.path.exists(os.path.join(CWD_PATH, "settings_app.py")):
                        print(_("Application pack: '%s' does not exists") % prj.strip())
                        return (None, None)
                    wx.CallAfter(app.run_script, app_name2, args[0])
                else:
                    print(_("Name of script: '%s' is not valid") % prg_name)
                    return (None, None)
        else:
            arg = args[0].strip()
            if arg == "embeded" or "." in arg or "/" in arg:
                if arg != "embeded":
                    CWD_PATH = os.path.join(PATHS["PRJ_PATH_ALT"], "_schremote")

                tmp = arg.replace("//", "$$$")
                if "/" in arg:
                    x = tmp.split("/", 1)
                    address = x[0].replace("$$$", "//")
                    if len(x) > 1:
                        app_name = x[1]
                        if app_name.endswith("/"):
                            app_name = app_name[:-1]
                else:
                    address = arg

                extern_prj = True
            else:
                CWD_PATH = os.path.join(PATHS["PRJ_PATH"], arg)
                if not os.path.exists(os.path.join(CWD_PATH, "settings_app.py")):
                    print(_("Application pack: '%s' does not exists") % arg)
                    return (None, None)
    sys.path.insert(0, CWD_PATH)

    httpclient.init_embeded_django()

    if not ("channels" in _PARAM or "rpc" in _PARAM):
        wx.Yield()
    import settings_app

    os.environ["DJANGO_SETTINGS_MODULE"] = "settings_app"
    from django.conf import settings

    if wx.Font.CanUsePrivateFont():
        fonts_path = os.path.join(settings.STATIC_ROOT, "fonts")
        font_names = [
            "DejaVuSansCondensed.ttf",
            "DejaVuSansCondensed-Bold.ttf",
            "DejaVuSansCondensed-Oblique.ttf",
            "DejaVuSansCondensed-BoldOblique.ttf",
            "DejaVuSerifCondensed.ttf",
            "DejaVuSerifCondensed-Bold.ttf",
            "DejaVuSerifCondensed-Italic.ttf",
            "DejaVuSerifCondensed-BoldItalic.ttf",
            "DejaVuSansMono.ttf",
            "DejaVuSansMono-Bold.ttf",
            "DejaVuSansMono-Oblique.ttf",
            "DejaVuSansMono-BoldOblique.ttf",
        ]
        for font_name in font_names:
            wx.Font.AddPrivateFont(os.path.join(fonts_path, font_name))

    if "sync" in _PARAM:
        from django.core.management.commands.migrate import Command as migrate_command

        migrate = migrate_command()
        try:
            migrate.run_from_argv(["manage.py", "migrate"])
        except SystemExit:
            pass
    if "loaddb" in _PARAM:
        from django.core.management.commands.loaddata import Command as load_command

        load = load_command()
        try:
            load.run_from_argv(["manage.py", "loaddata"])
        except:
            pass

    cwd = CWD_PATH
    inst_dir = SRC_PATH
    if inst_dir == "":
        inst_dir = cwd

    settings.TEMPLATES[0]["DIRS"].insert(0, inst_dir + "/appdata/plugins")
    settings.TEMPLATES[0]["DIRS"].insert(0, cwd + "/appdata/plugins")
    settings.TEMPLATES[0]["DIRS"].insert(0, inst_dir + "/../templates")
    settings.TEMPLATES[0]["DIRS"].insert(0, cwd + "/templates")

    for a in apps:
        settings.INSTALLED_APPS.append(a)

    port = 0
    if "server_only" in _PARAM:
        port = 8000
        if ":" in address:
            l = address.split(":")
            if len(l) == 3:
                port = int(l[2])
        address = "embeded"
    if address == "embeded":
        import socket
        from pytigon_lib.schdjangoext.server import run_server

        if "server_only" in _PARAM:
            address = "127.0.0.1"
            if "listen" in _PARAM:
                if ":" in _PARAM["listen"]:
                    address, port = _PARAM["listen"].split(":")
                    port = int(port)
                else:
                    address = _PARAM["listen"]
                    port = 8000
        else:
            address = "127.0.0.3"
        if port == 0:
            port = 8000
        test = True
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while test:
            try:
                s.bind((address, port))
                s.close()
                s = None
                test = False
            except:
                port += 1

        if "extra" in _PARAM:
            server = run_server(address, port, prod=False, params=_PARAM["extra"])
        else:
            server = run_server(address, port, prod=False)
        address = "http://" + address + ":" + str(port)
    else:
        #    from pytigon_lib.schtasks.base_task import get_process_manager
        #    app.task_manager = get_process_manager()
        server = None

    if "embeded_taskqueue" in _PARAM:
        app.task_manager = ""
        from django_q.management.commands.qcluster import Command as qcluster_command

        qcluster = qcluster_command()
        # try:
        if True:
            # qcluster.run_from_argv(["manage.py", "qcluster"])
            app.task_manager = Process(
                target=qcluster.run_from_argv, args=(["manage.py", "qcluster"],)
            )
            app.task_manager.start()
            print("Task manager started")
            # p.join()
        # except SystemExit:
        #    pass

    settings.BASE_URL = "http://" + address
    settings.URL_ROOT_FOLDER = ""
    if not "server_only" in _PARAM:
        init_ret = app._init2(address, app_name)
        if init_ret != 200:
            return (False, False)

    if app.authorized:
        reinit = False
    else:
        reinit = True

    app.server = server
    app.cwd = cwd
    app.inst_dir = inst_dir

    tab = app.get_tab(0)

    app.title = app_title
    autologin = True
    if "server_only" in _PARAM:
        app.gui_style = "app.gui_style = tray(file(exit,open))"
        app.authorized = True
        reinit = False
    else:
        for row in tab:
            if row[0].data == "autologin":
                if row[1].data == "1":
                    autologin = True
                else:
                    autologin = False
            elif row[0].data == "gui_style":
                app.gui_style = row[1].data
            elif row[0].data == "csrf_token":
                app.csrf_token = row[1].data
            elif "start_page" in row[0].data:
                app.start_pages.extend(
                    [x for x in row[1].data.split(";") if x and x != "None"]
                )
            elif row[0].data == "title":
                app.title = row[1].data
            elif row[0].data == "plugins":
                if row[1].data and row[1].data != "":
                    app.plugins = row[1].data.split(";")

    app._install_plugins()

    ready_to_run = True

    if not app.authorized and (
        (autologin and not "username" in _PARAM)
        or ("username" in _PARAM and "password" in _PARAM)
    ):
        if "username" in _PARAM:
            username2 = _PARAM["username"]
            password2 = _PARAM["password"]
        else:
            username2 = "auto"
            password2 = "anawa"

        ready_to_run = False
        response = app.http.post(
            app,
            "/" + app_name + "/schsys/do_login/?from_pytigon=1"
            if app_name
            else "/schsys/do_login/?from_pytigon",
            {
                #'csrfmiddlewaretoken': app.csrf_token,
                "username": username2,
                "password": password2,
                "next": address + "/" + app_name + "/schsys/ok/"
                if app_name
                else address + "/schsys/ok/",
                "client_param": app._get_parm_for_server(),
            },
        )
        ret_str = response.str()
        if "$$RETURN_OK" in ret_str:
            app.authorized = True
            ready_to_run = True
        else:
            print("Error: %s" % ret_str)
    if not app.authorized and "username" in _PARAM:
        ready_to_run = False
        href = "/" + app_name + "/" if app_name else "/"
        if login(href, auth_type=None, username=_PARAM["username"]):
            app.authorized = True
            ready_to_run = True
    if reinit:
        app._re_init(address, app_name)
    return (ready_to_run, True if "nogui" in _PARAM else False)


def _main_run():
    app = wx.GetApp()
    app.locale = wx.Locale(wx.LANGUAGE_DEFAULT)
    app.locale.AddCatalogLookupPathPrefix(SRC_PATH + "/pytigon_gui/locale")

    app.locale.AddCatalog("wx")
    app.locale.AddCatalog("pytigon")

    if app.embeded_browser:
        frame = browserframe.SchBrowserFrame(
            None,
            app.gui_style,
            wx.ID_ANY,
            app.title,
            wx.DefaultPosition,
            wx.Size(_APP_SIZE[0], _APP_SIZE[1]),
        )
    else:
        frame = appframe.SchAppFrame(
            app.gui_style,
            app.title,
            wx.DefaultPosition,
            wx.Size(_APP_SIZE[0], _APP_SIZE[1]),
            video_name=_VIDEO_NAME,
        )

    frame.CenterOnScreen()

    # if not "tray" in app.gui_style:
    #    frame.Show()

    wx.Log.SetActiveTarget(wx.LogStderr())

    destroy_fun_tab = frame.destroy_fun_tab
    httpclient.set_http_error_func(http_error)

    def idle_fun():
        wx.GetApp().web_ctrl.OnIdle(None)

    httpclient.set_http_idle_func(idle_fun)

    if _RPC:
        reactor.listenTCP(app.rpc, server.Site(app))

    if _WEBSOCKET and app.base_address:
        if ";" in _WEBSOCKET:
            websockets = _WEBSOCKET.split(";")
        else:
            websockets = [_WEBSOCKET]

        local = (
            True
            if app.base_address and app.base_address.startswith("http://127.0.0.2")
            else False
        )

        for websocket_id in websockets:
            create_websocket_client(app, websocket_id, local)

    if "import" in _PARAM:
        x = __import__(_PARAM["import"])

        def s():
            nonlocal frame, x
            x(frame)

        wx.CallAfter(s)

    if hasattr(wx, "pseudoimport"):
        x = getattr(wx, "pseudoimport")

        def s():
            nonlocal frame, x
            x(frame)

        wx.CallAfter(s)

    if "channels" in _PARAM or "rpc" in _PARAM or "websocket" in _PARAM:
        LOOP.run_until_complete(app.MainLoop())
    else:
        app.MainLoop()

    if app.server:
        app.server.stop()
    del app
    for pos in destroy_fun_tab:
        pos()


def main():
    ready_to_run, nogui = _main_init()
    if ready_to_run:
        if nogui:
            while True:
                time.sleep(100)
        else:
            _main_run()
    os._exit(0)


if __name__ == "__main__":
    main()



"""This module contains SchNotebookPage. It represent one tab window in SchNotebook. One SchNotebookPage can manage
multiple SchPage windows.
"""

import wx
from pytigon_gui.guiframe import page


class SchNotebookPage(wx.Window):
    """SchNotebookPage represent one tab pytigon applications"""

    def __init__(self, parent):
        """Constructor"""
        wx.Window.__init__(
            self,
            parent,
            style=wx.TAB_TRAVERSAL | wx.WANTS_CHARS,
            name="SchNotebookPage",
        )
        try:
            self.SetBackgroundStyle(wx.BG_STYLE_ERASE)
        except:
            self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self._start_pos = None
        self._start_pos_x_y_dx_dy = None
        self._last_x_y_dx_dy = None
        self._best_x_y_dx_dy = None
        self._layout_style = 0

        self.child_panels = []
        self.bestx = -1
        self.orient = None
        self.http = None
        self.reverse_style = True

        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.on_erase_background)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LEFT_UP, self.on_left_up)
        self.Bind(wx.EVT_MOTION, self.on_move)
        self.Bind(wx.EVT_SET_FOCUS, self.on_set_focus)

        self.SetWindowStyleFlag(wx.WANTS_CHARS)

    def on_set_focus(self, evt):
        if self.get_page_count() > 0:
            self.get_page(-1).SetFocus()

    def get_app_http(self):
        """Returns  :class:`~pytigon_lib.schhttptools.httpclient.HttpClient` object connected to this object"""
        return self.http

    def on_erase_background(self, evt):
        if not wx.GetApp().GetTopWindow():
            return
        dc = wx.ClientDC(self)
        dc.Clear()
        margin = self.get_margins()
        if hasattr(wx.GetApp().GetTopWindow(), "desktop"):
            tabs_count = len(wx.GetApp().GetTopWindow().desktop._mgr.GetAllPanes())
            if (
                self.get_page_count()
                and tabs_count > 2
                or self.get_page_count() > 1
                or wx.GetApp().GetTopWindow().count_shown_panels(count_toolbars=False)
                > 1
                and self.get_page_count() > 0
            ):
                (dx, dy) = self.get_page(-1).GetSize()
                (x, y) = self.get_page(-1).GetPosition()
                x = x - margin / 2
                y = y - margin / 2
                dx = dx + margin
                dy = dy + margin
                if (
                    self.GetParent().active
                    and self.GetParent().GetCurrentPage() == self
                ):
                    col = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
                else:
                    col = wx.SystemSettings.GetColour(wx.SYS_COLOUR_INACTIVEBORDER)
                dc.SetPen(wx.Pen(col, margin))
                dc.DrawLine(int(x), int(y), int(x + dx), int(y))
                dc.DrawLine(int(x + dx), int(y), int(x + dx), int(y + dy))
                dc.DrawLine(int(x), int(y - margin / 2), int(x), int(y + dy))
                dc.DrawLine(int(x), int(y + dy), int(x + dx), int(y + dy))
                if self._layout_style > 0:
                    col = wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)
                    dc.SetPen(wx.Pen(col, 1))
                    if self._layout_style in (1, 3):
                        dc.DrawLine(
                            int(x - margin), int(y), int(x - margin), int(y + dy)
                        )
                        dc.DrawLine(
                            int(x - 5 * margin),
                            int(y),
                            int(x - 5 * margin),
                            int(y + dy),
                        )
                        for i in range(-4, 5, 2):
                            dc.DrawLine(
                                int(x - 5 * margin + 2),
                                int(y + dy / 2 + margin * i),
                                int((x - margin) - 1),
                                int(y + dy / 2 + margin * i),
                            )
                    if self._layout_style in (2, 3):
                        dc.DrawLine(
                            int(x), int(y - margin), int(x + dx), int(y - margin)
                        )
                        dc.DrawLine(
                            int(x),
                            int(y - 5 * margin),
                            int(x + dx),
                            int(y - 5 * margin),
                        )
                        for i in range(-4, 5, 2):
                            dc.DrawLine(
                                int(x + dx / 2 + margin * i),
                                int(y - 5 * margin + 2),
                                int(x + dx / 2 + margin * i),
                                int((y - margin) - 1),
                            )

    def on_left_down(self, event):
        self._start_pos = event.GetPosition()
        self._start_pos_x_y_dx_dy = self._last_x_y_dx_dy
        self.SetCursor(wx.Cursor(wx.CURSOR_SIZING))
        self.CaptureMouse()
        event.Skip()

    def on_left_up(self, event):
        if self._start_pos:
            self._start_pos = None
            self._start_pos_x_y_dx_dy = None
            self._best_x_y_dx_dy = None
            self.ReleaseMouse()
            self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
        event.Skip()

    def on_move(self, event):
        if self._start_pos and self._start_pos_x_y_dx_dy:
            pos = event.GetPosition()
            dx = pos[0] - self._start_pos[0]
            dy = pos[1] - self._start_pos[1]
            if self.reverse_style:
                dy = -1 * dy
            self._best_x_y_dx_dy = (
                self._start_pos_x_y_dx_dy[0] + dx,
                self._start_pos_x_y_dx_dy[1] + dy,
                self._start_pos_x_y_dx_dy[2],
                self._start_pos_x_y_dx_dy[3],
            )
            self._layout()
        event.Skip()

    def get_xy(self, size=None, page=-1):
        """get best point whitch divide window for place new SchPage window"""
        if size == None:
            size = self.GetSize()
        if self._best_x_y_dx_dy:
            return self._best_x_y_dx_dy
        else:
            (bestx, besty) = self.get_page(page).calculate_best_size()
            self.bestx = bestx
            self.besty = besty
        if bestx > size.GetWidth() - 20:
            x = size.GetWidth() / 10
        else:
            x = (size.GetWidth() - 2) - bestx
        if besty > size.GetHeight() - 20:
            y = size.GetHeight() / 10
        else:
            y = (size.GetHeight() - 2) - besty
        self._last_x_y_dx_dy = (x, y, size.GetWidth(), size.GetHeight())
        return self._last_x_y_dx_dy

    def get_margins(self):
        """get margin size between child windows"""
        return 2

    def _set_dimensions(self, page, x, y, width, height, dx, dy):
        if x > 0 and y > 0 and width >= 0 and height >= 0:
            if self.child_panels[-1].vertical_position:
                if self.child_panels[-1].vertical_position == "top":
                    page.SetSize(int(x), int(dy - height - y), int(width), int(height))
                else:
                    page.SetSize(int(x), int(y), int(width), int(height))
            else:
                if self.reverse_style:
                    page.SetSize(int(x), int(dy - height - y), int(width), int(height))
                else:
                    page.SetSize(int(x), int(y), int(width), int(height))

    def _layout(self, size=None):
        if self.get_page_count() > 0:
            count = self.get_page_count()
            margin = self.get_margins()
            if count == 0:
                pass
            if count == 1:
                self._layout_style = 0
                if not size:
                    size = self.GetSize()
                if size:
                    dx = size.GetWidth()
                    dy = size.GetHeight()
                self._set_dimensions(
                    self.get_page(0),
                    margin,
                    margin,
                    dx - 2 * margin,
                    dy - 2 * margin,
                    dx,
                    dy,
                )
            if count > 1:
                (x, y, dx, dy) = self.get_xy(size)
                if not self._best_x_y_dx_dy:
                    if self.bestx >= 0 and x > self.bestx:
                        x = self.bestx
                    else:
                        self.bestx = x
                else:
                    self._last_x_y_dx_dy = self._best_x_y_dx_dy
                if count == 2:
                    if (dx - x) * dy < 2 * ((dy - y) * dx):
                        self._layout_style = 1
                        self._set_dimensions(
                            self.get_page(0),
                            margin,
                            margin,
                            x - 2 * margin,
                            dy - 2 * margin,
                            dx,
                            dy,
                        )
                        self._set_dimensions(
                            self.get_page(1),
                            x + 5 * margin,
                            margin,
                            (dx - x) - 6 * margin,
                            dy - 2 * margin,
                            dx,
                            dy,
                        )
                    else:
                        self._layout_style = 2
                        self._set_dimensions(
                            self.get_page(0),
                            margin,
                            margin,
                            dx - 2 * margin,
                            y - 2 * margin,
                            dx,
                            dy,
                        )
                        self._set_dimensions(
                            self.get_page(1),
                            margin,
                            y + 5 * margin,
                            dx - 2 * margin,
                            (dy - y) - 6 * margin,
                            dx,
                            dy,
                        )
                if count > 2:
                    self._layout_style = 3
                    for i in range(0, count - 2):
                        self._set_dimensions(
                            self.get_page(i),
                            margin,
                            margin + (i * dy) / (count - 2),
                            x - 2 * margin,
                            dy / (count - 2) - 2 * margin,
                            dx,
                            dy,
                        )
                    self._set_dimensions(
                        self.get_page(-2),
                        x + margin,
                        margin,
                        (dx - x) - 2 * margin,
                        y - 2 * margin,
                        dx,
                        dy,
                    )
                    self._set_dimensions(
                        self.get_page(-1),
                        x + 5 * margin,
                        y + 5 * margin,
                        (dx - x) - 6 * margin,
                        (dy - y) - 6 * margin,
                        dx,
                        dy,
                    )
            self.Refresh()

    def get_page_count(self):
        """Get number of child schFrameForm objects"""
        return len(self.child_panels)

    def get_page(self, nr):
        """return child SchPage object

        Args:
            nr: number of page
        """
        return self.child_panels[nr]

    def add_page(self, page):
        """Append child SchPage object

        Args:
            page - SchPage object
        """
        if self.get_page_count() > 0:
            if page.disable_parent:
                self.get_page(-1).enable_forms(False)
        self.child_panels.append(page)
        self._layout()

    def on_size(self, event):
        size = event.GetSize()
        self.bestx = -1
        # self._layout(size)
        if size:
            wx.CallAfter(self._layout, size)
        event.Skip()

    def close_no_del(self, close_without_refresh=True):
        """Close last in hierarchy child form.

        Args:
            close_without_refresh - if True parent form of closing form shoud be refreshed,
            if False refresh is not needed.
        """
        ret = 0
        count = self.get_page_count()
        if count > 0:
            win = self.get_page(-1)
            if win.CanClose():
                del self.child_panels[-1]
                if count > 1:
                    self.get_page(-1).enable_forms(True)
                    self._layout()
                    self.get_page(-1).set_focus()
                    if close_without_refresh:
                        self.get_page(-1).signal("child_canceled", win)
                    else:
                        self.get_page(-1).signal("child_closed_with_ok", win)
                    ret = True
                else:
                    ret = False
                win.Destroy()
                return ret
            else:
                return True

    def close_child_page(self, close_without_refresh=True):
        """Close last in hierarchy child form. If there is only one form, besides closing form,
        this SchNotebookPage is also closed

        Args:
            close_without_refresh - if True parent form of closing form shoud be refreshed,
            if False refresh is not needed.
        """
        pages = self.GetParent()._tabs._pages
        for i in range(0, len(pages)):
            if pages[i].window == self:
                ret = self.close_no_del(close_without_refresh)
                if not ret:
                    self.GetParent().DeletePage(i)
                return ret
        return False

        # sel = self.GetParent().GetSelection()
        # if sel>=0:
        #    ret = self.close_no_del(close_without_refresh)
        #    if not ret:
        #        p = self.GetParent()
        #        p.DeletePage(sel)
        #    return ret
        # else:
        #    return False

    def on_child_form_ok(self):
        """function called by child SchForm object when OK is pushed and data in parent form shoud be refreshed"""
        self.close_child_page(False)

    def on_child_form_cancel(self):
        """function called by child SchForm object when Cancel is pushed and data in parent form doesn't need be refreshed"""
        self.close_child_page(True)

    # def select_tab(self):
    #    n = self.GetParent()
    #    src_idx = self.GetParent()._tabs.GetIdxFromWindow(self)
    #    sel = n.GetSelection()
    #    if sel != src_idx:
    #        n.SetSelection(src_idx)

    def activate_page(self):
        if self.get_page_count() > 0:
            self.get_page(-1).activate_page()
        else:
            self.SetFocus()

    def deactivate_page(self):
        if self.get_page_count() > 0:
            self.get_page(-1).deactivate_page()

    # def refresh_html(self):
    #    pass

    def new_child_page(
        self, address_or_parser, title="", parameters=None, callback=None
    ):
        """Append a new page

        Args:
            address_or_parser: can be: address of http page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'

            title - new page title
            parameters - dict
        """
        h = page.SchPage(self, address_or_parser, parameters)
        if self.get_page_count() > 0:
            h.parent_page = self.get_page(-1)
        nr = 0
        if h.header != None:
            nr = nr + 4
        if h.footer != None:
            nr = nr + 2
        if h.panel != None:
            nr = nr + 1
        if title and title != "":
            title2 = title
        else:
            title2 = h.get_title()
        self.add_page(h)  # , title2, True, nr)

        def init_page():
            nonlocal h, callback
            h.init_frame(callback)
            h.activate_page()
            wx.GetApp().GetTopWindow()._mgr.GetPane("desktop").Show()
            h.Update()
            # if callback:
            #    return callback(h)

        wx.CallAfter(init_page)
        return h

    def new_main_page(self, address_or_parser, title="", parameters=None, view_in=None):
        """Create new top

        Args:
            address_or_parser: can be: address of http page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'

            title - new page title

            parameters - dict

            view_in - can be: 'desktop', 'panel', 'header' or 'footer'
        """
        if view_in == None:
            pp = self.GetParent().GetParent().GetParent()._mgr.GetPane(self.GetParent())
            return (
                wx.GetApp()
                .GetTopWindow()
                .new_main_page(address_or_parser, title, parameters, pp.name)
            )
        else:
            return (
                wx.GetApp()
                .GetTopWindow()
                .new_main_page(address_or_parser, title, parameters, view_in)
            )



"""SchPage is a container for 4 SchForm classes: body, header, footer and panel. Only "body" object is obligatory."""

import wx

try:
    from wx.adv import LayoutAlgorithm
except:
    from wx import LayoutAlgorithm
from pytigon_gui.guilib import events
from pytigon_gui.guilib.signal import Signal
from pytigon_lib.schtools import createparm
from pytigon_lib.schparser.html_parsers import ShtmlParser
from pytigon_lib.schhttptools.httpclient import HttpResponse
from pytigon_gui.guiframe.form import SchForm
from pytigon_gui.guilib.events import *


class SchPage(wx.Window, Signal):
    """SchPage class"""

    def __init__(
        self, parent, address_or_parser, parameters, pos=(0, 0), size=wx.DefaultSize
    ):
        """Contructor

        Args:
            parent - parent window
            address_or_parser: can be: address of http page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'
            parameters: dict
        """

        self._active = True
        self._ctrl_dict_old = {}
        self._ctrl_dict = {}
        self._last_size = None
        self._disable_setfocus = 0
        self._signal_handlers = []

        self.address_or_parser = address_or_parser
        self.parameters = parameters
        self.default_button = None
        self.parent_page = None
        self.last_control_with_focus = None
        self.vertical_position = None

        self.header = None
        self.panel = None
        self.body = None
        self.footer = None
        self.active_form = None
        self.active_ctrl = None

        Signal.__init__(self)
        wx.Window.__init__(
            self, parent, -1, pos, size, style=wx.WANTS_CHARS, name="SchPage"
        )

        self.exists = True

        mp = self._read_html(address_or_parser, parameters)
        if not mp:
            print("ERROR READ HTML:", address_or_parser, parameters)

        self.address = mp.address

        header = mp.get_header()
        body = mp.get_body()
        footer = mp.get_footer()
        panel = mp.get_panel()
        config = mp.var

        vscroll = True
        hscroll = True

        if "no_vscrollbar" in config:
            vscroll = False
        if "no_hscrollbar" in config:
            hscroll = False
        if "vertical_position" in config:
            self.set_vertical_position(config["vertical_position"])

        self.title = mp.title

        if "disable_parent" in config:
            if config["disable_parent"] == "0":
                self.disable_parent = False
            else:
                self.disable_parent = True
        else:
            self.disable_parent = True

        winids = []

        if header[0]:
            topwin = wx.adv.SashLayoutWindow(
                self,
                -1,
                wx.DefaultPosition,
                (100, 5),
                wx.adv.SW_3D,
                style=wx.WANTS_CHARS,
            )
            topwin.SetOrientation(wx.adv.LAYOUT_HORIZONTAL)
            topwin.SetAlignment(wx.adv.LAYOUT_TOP)
            topwin.SetSashVisible(wx.adv.SASH_BOTTOM, True)
            self.header = SchForm(topwin, page=self, form_type="header")
            self.header.show_form(header)
            dy = self.header.calculate_best_size()[1]
            topwin.SetDefaultSize((1000, dy))
            self._top_window = topwin
            winids.append(topwin.GetId())
        else:
            self._top_window = None

        if footer[0]:
            bottomwin = wx.adv.SashLayoutWindow(
                self, -1, wx.DefaultPosition, (1000, 5), wx.adv.SW_3D
            )
            bottomwin.SetOrientation(wx.adv.LAYOUT_HORIZONTAL)
            bottomwin.SetAlignment(wx.adv.LAYOUT_BOTTOM)
            bottomwin.SetSashVisible(wx.adv.SASH_TOP, True)
            self.footer = SchForm(bottomwin, page=self, form_type="footer")
            self.footer.show_form(footer)
            bottomwin.SetDefaultSize((1000, self.footer.calculate_best_size()[1]))
            self._bottom_window = bottomwin
            winids.append(bottomwin.GetId())
        else:
            self._bottom_window = None

        if panel[0]:
            leftwin = wx.adv.SashLayoutWindow(
                self, -1, wx.DefaultPosition, (5, 100), wx.adv.SW_3D
            )
            leftwin.SetOrientation(wx.adv.LAYOUT_VERTICAL)
            leftwin.SetAlignment(wx.adv.LAYOUT_LEFT)
            leftwin.SetSashVisible(wx.adv.SASH_RIGHT, True)
            self.panel = SchForm(leftwin, page=self, form_type="panel")
            self.panel.show_form(panel)
            xy = self.panel.calculate_best_size()
            leftwin.SetDefaultSize(xy)
            self._left_window = leftwin
            winids.append(leftwin.GetId())
        else:
            self._left_window = None

        self.body = SchForm(self, self, hscroll, vscroll)
        self.body.set_htm_type("body")
        attrs = mp.get_body_attrs()
        if "width" in attrs and "height" in attrs:
            w = attrs["width"]
            h = attrs["height"]
            self.body.bestsize = (int(w) * 4 / 3, int(h) * 4 / 3)
        self.body.show_form(body, parameters)
        self.body.set_address_parm(self.address)
        if "refresh" in mp.var:
            self.body.refresh_time(int(mp.var["refresh"]))

        if winids:
            self.Bind(
                wx.adv.EVT_SASH_DRAGGED_RANGE,
                self.on_sash_drag,
                id=min(winids),
                id2=max(winids),
            )

        self.Bind(wx.EVT_SIZE, self.on_size)
        self.body.Bind(wx.EVT_CHILD_FOCUS, self.on_child_focus)
        self.Bind(wx.EVT_SET_FOCUS, self.on_set_focus)

        self.register_signal(self, "child_closed_with_ok")

    def child_closed_with_ok(self, win=None):
        wx.CallAfter(self.refresh_html)

    def reg_application_signal_handler(self, fun, signal):
        """register callback function for signal

        Args:
            fun - callback function
            signal - name of signal
        """
        for pos in self._signal_handlers:
            if pos[1] == signal:
                break
        dispatcher.connect(fun, signal, sender=dispatcher.Any)
        self._signal_handlers.append((fun, signal))

    def unreg_application_signal_handler(self, signal):
        """Unregister signal" """
        i = 0
        test = None
        for pos in self._signal_handlers:
            if pos[1] == signal:
                test = pos
                break
            i += 1
        if test:
            dispatcher.disconnect(pos[0], pos[1], sender=dispatcher.Any)
            del self._signal_handlers[i]

    def _read_html(self, address_or_parser, parameters):
        """Run a http request and read result page

        Args:
            address_or_parser: can be: address of http page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'
            parameters: dict
        """
        mp, adr = wx.GetApp().read_html(self, address_or_parser, parameters)
        return mp

    def append_ctrl(self, obj):
        self._ctrl_dict[obj.get_unique_name()] = obj

    def restart_ctrl_lp(self):
        self._ctrl_dict_old = self._ctrl_dict
        self._ctrl_dict = {}

    def pop_ctrl(self, name):
        if name in self._ctrl_dict_old:
            ret = self._ctrl_dict_old[name]
            del self._ctrl_dict_old[name]
            return ret
        else:
            return None

    def test_ctrl(self, name):
        if name in self._ctrl_dict:
            return True
        else:
            return False

    def remove_old_ctrls(self):
        for name in self._ctrl_dict_old:
            win = self._ctrl_dict_old[name]
            win.Destroy()
            # if name in self.body.controls:
            #   del self.body.controls[name]
        self._ctrl_dict_old = {}

    def __getitem__(self, key):
        return self._ctrl_dict[key]

    def get_widgets(self):
        return self._ctrl_dict

    def init_frame(self, callback=None):
        if self.header:
            self.header.init()
        if self.footer:
            self.footer.init()
        if self.panel:
            self.panel.init()

        self.body.init(callback)

    def get_parent_page(self):
        """Return parent wxPage object"""
        return self.parent_page

    def set_default_button(self, button):
        """Set default button, when Enter key is pressed action connected to button is invoked"""
        self.default_button = button

    def on_key_down(self, event):
        if event.KeyCode == wx.WXK_RETURN and self.default_button:
            self.default_button.OnClick(event)
        else:
            event.Skip()

    def on_char_hook(self, event):
        if event.KeyCode == wx.WXK_ESCAPE:
            if hasattr(self.GetParent(), "on_child_form_cancel"):
                self.GetParent().on_child_form_cancel()
            else:
                event.Skip()
        else:
            event.Skip()

    def get_title(self):
        """Return page title"""
        return self.title

    def set_adr_and_param(self, address_or_parser, param):
        """Modify web address and parameters

        Args:
            address_or_parser: can be: address of http page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'
            parameters: dict
        """
        self.address_or_parser = address_or_parser
        self.parameters = param

    def _refresh(self):
        if self.address_or_parser.__class__.__name__ == "ShtmlParser":
            if self.address_or_parser.address:
                mp = self._read_html(self.address_or_parser.address, self.parameters)
            else:
                mp = self._read_html(self.address_or_parser, self.parameters)
        else:
            if type(self.address_or_parser) == HttpResponse:
                mp = self._read_html(self.address_or_parser.url, self.parameters)
            else:
                mp = self._read_html(self.address_or_parser, self.parameters)
        if not mp:
            return
        address = mp.address
        header = mp.get_header()
        body = mp.get_body()
        footer = mp.get_footer()
        panel = mp.get_panel()
        config = mp.var
        self.title = mp.title
        if self.header:
            if not self.header.show_form(
                """<html encoding="utf-8">""" + header + "</html>"
            ):
                return False
        if self.body:
            if not self.body.show_form(body, self.parameters):
                return False
        if self.footer:
            if not self.footer.show_form(
                """<html encoding="utf-8">""" + footer + "</html>"
            ):
                return False
        if self.panel:
            if not self.panel.show_form(
                """<html encoding="utf-8">""" + panel + "</html>"
            ):
                return False
        if self.header:
            self.header.init()
        if self.footer:
            self.footer.init()
        if self.panel:
            self.panel.init()

        self.body.Refresh()
        self.body.init()
        return True

    def CanClose(self):
        if self.body:
            if not self.body.CanClose():
                return False
        if self.header:
            if not self.header.CanClose():
                return False
        if self.footer:
            if not self.footer.CanClose():
                return False
        if self.panel:
            if not self.panel.CanClose():
                return False

        if self.header:
            self.header._on_close()
        if self.body:
            self.body._on_close()
        if self.footer:
            self.footer._on_close()
        if self.panel:
            self.panel._on_close()

        for pos in self._signal_handlers:
            dispatcher.disconnect(pos[0], pos[1], sender=dispatcher.Any)

        return True

    def on_set_focus(self, evt):
        if self.body:
            self.body.SetFocus()

    def on_child_focus(self, evt):
        if not self._disable_setfocus:
            new_win = evt.GetWindow()
            parent = new_win.GetParent()
            while parent != None:
                if parent.__class__.__name__ == "HtmlPanel":
                    parent.SelectTab()
                    break
                parent = parent.GetParent()
            while new_win != None:
                if (
                    new_win.GetParent()
                    and new_win.GetParent().GetWindowStyleFlag() & wx.TAB_TRAVERSAL != 0
                ):
                    break
                new_win = new_win.GetParent()
            if new_win != self.last_control_with_focus:
                if self.last_control_with_focus:
                    if hasattr(self.last_control_with_focus, "KillFocus"):
                        self._disable_setfocus = True
                        getattr(self.last_control_with_focus, "KillFocus")()
                        self._disable_setfocus = False
                self.last_control_with_focus = new_win
            evt.Skip()
        else:
            evt.Skip()

    def set_focus(self):
        if self.last_control_with_focus:
            self.last_control_with_focus.SetFocus()
        else:
            super().SetFocus()

    def on_sash_drag(self, event):
        if event.GetDragStatus() == wx.SASH_STATUS_OUT_OF_RANGE:
            return
        eobj = event.GetEventObject()
        if eobj is self._top_window:
            self._top_window.SetDefaultSize((1000, event.GetDragRect().height))
        elif eobj is self._left_window:
            self._left_window.SetDefaultSize((event.GetDragRect().width, 1000))
        elif eobj is self.bottom_window:
            self._bottom_window.SetDefaultSize((1000, event.GetDragRect().height))
        LayoutAlgorithm().LayoutWindow(self, self.body)
        self.body.Refresh()

    def on_size(self, event):
        LayoutAlgorithm().LayoutWindow(self, self.body)
        self.body.Refresh()
        if event:
            event.Skip()

    def get_last_control_with_focus(self):
        return self.last_control_with_focus

    def refresh_html(self, method=0):
        """Reload page from the server"""
        return self._refresh_html(method)

    def _refresh_html(self, method=0):
        ret = self._refresh()
        self.body.wxdc = None
        self.body.update_controls = True
        self.body.Refresh()
        if not ret:
            return False
        return True

    def has_parm(self, param):
        if param in self._ctrl_dict:
            return True
        return self.item_exist(param)

    def get_parm(self, param):
        if param in self._ctrl_dict:
            ctrl = self._ctrl_dict[param]
        else:
            ctrl = self.get_item(param)
        if ctrl:
            return ctrl.GetValue()
        else:
            return None

    def get_item(self, ctrl_name):
        if ctrl_name in self._ctrl_dict:
            return self._ctrl_dict[ctrl_name]
        if self.body != None and hasattr(self.body, ctrl_name):
            return getattr(self.body, ctrl_name)
        if self.header != None and hasattr(self.header, ctrl_name):
            return getattr(self.header, ctrl_name)
        if self.panel != None and hasattr(self.panel, ctrl_name):
            return getattr(self.panel, ctrl_name)
        if self.footer != None and hasattr(self.footer, ctrl_name):
            return getattr(self.footer, ctrl_name)
        return None

    def item_exist(self, ctrl_name):
        """Test if a page have specified control

        Args:
            ctrl_name -  name of control element
        """
        if ctrl_name in self._ctrl_dict:
            return True
        if self.body and hasattr(self.body, ctrl_name):
            return True
        if self.header and hasattr(self.header, ctrl_name):
            return True
        if self.panel and hasattr(self.panel, ctrl_name):
            return True
        if self.footer and hasattr(self.footer, ctrl_name):
            return True
        return False

    def __getitem__(self, key):
        return self.get_item(key)

    def calculate_best_size(self):
        if self._last_size:
            dx = self._last_size.GetWidth()
            dy = self._last_size.GetHeight()
            self._last_size = None
            return (dx, dy)
        dx = 0
        dy = 0

        if self.header:
            (x, y) = self.header.calculate_best_size()
            dy = dy + y
        if self.body:
            (x, y) = self.body.calculate_best_size()
            dx = dx + x
            dy = dy + y
        if self.footer:
            (x, y) = self.footer.calculate_best_size()
            dy = dy + y
        if self.panel:
            (x, y) = self.panel.calculate_best_size()
            dx = dx + x
            if y > dy:
                dy = y
        return (dx, dy)

    def enable_forms(self, enable):
        """Enable or disable managed by this window forms

        Params:
            enable - if True enable forms, if False - disable forms
        """
        if enable == False:
            self._last_size = self.GetSize()
        if not (enable == False and not self.disable_parent):
            if self.header:
                self.header.enable(enable)
            if self.body:
                self.body.enable(enable)
            if self.footer:
                self.footer.enable(enable)
            if self.panel:
                self.panel.enable(enable)

    def change_notebook_page_title(self, new_title):
        """Change a title of notebook witch contain this SchPage object"""
        p = self.GetParent()
        tab = p.GetParent()
        sel = tab.GetPageIndex(p)
        tab.SetPageText(sel, new_title)

    def set_new_href(self, href):
        """Set new web address"""
        self.address_or_parser = href

    def activate_page(self):
        self._active = True
        self.body.SetFocus()

    def deactivate_page(self):
        self._active = False

    def is_active(self):
        return self._active

    def set_page(self, page_source):
        """Set content of body form asigned to this page

        Args:
            page_source: new text content
        """
        self.body.set_page(page_source)

    def set_vertical_position(self, position):
        """Position of child form

        position:
            "top"
            "bottom"
            None or "default" (defautl)
        """
        self.vertical_position = position

    def close(self):
        self.GetParent().on_child_form_cancel()



from setuptools import setup, find_packages


def package_files(directory, ext=None):
    paths = []
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            if not ext or (ext and filename.endswith(ext)):
                paths.append(os.path.join("..", path, filename))
    return paths


extra_files = ["../requirements.txt"]


with open("requirements.txt") as f:
    tmp = f.read().strip().split("\n")
    install_requires = [pos for pos in tmp if "://" not in pos]
    dependency_links = [pos for pos in tmp if "://" in pos]


setup(
    name="pytigon-gui",
    version="0.250514",
    description="Pytigon GUI",
    author="Sławomir Chołaj",
    author_email="slawomir.cholaj@gmail.com",
    license="LGPLv3",
    packages=find_packages(),
    package_data={"": extra_files},
    install_requires=install_requires,
    dependency_links=dependency_links,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    entry_points={
        "gui_scripts": [
            "ptigw=pytigon.pytigon_run:run",
        ]
    },
)



"""
Top window
"""

from wx.lib.agw import aui
import wx


class SChAuiBaseManager(aui.framemanager.AuiManager):
    def __init__(self, *argi, **argv):
        aui.framemanager.AuiManager.__init__(self, *argi, **argv)
        self.Bind(wx.EVT_WINDOW_CREATE, self.DoUpdateEvt)

    # def Update(self):
    #    if '__WXGTK__' in wx.PlatformInfo:
    #        def _fun():
    #            self.DoUpdate()
    #            if self._frame:
    #                self._frame.Refresh()
    #        wx.CallAfter(_fun)
    #    else:
    #        super().Update()

    # def OnRender(self, event):
    #    if self._frame and self._frame.GetHandle():
    #        super().OnRender(event)
    #    else:
    #        event.Skip()

    def OnLeftDown(self, event):
        part = self.HitTest(*event.GetPosition())
        if not part.type in [0, 1]:
            super().OnLeftDown(event)


class SChAuiManager(SChAuiBaseManager):
    def __init__(self, *argi, **argv):
        aui.AuiManager.__init__(self, *argi, **argv)

    def AddPane(self, window, arg1, *argi, **argv):
        ret = aui.AuiManager.AddPane(self, window, arg1, *argi, **argv)
        if hasattr(window, "SetPanel"):
            window.SetPanel(arg1)
        return ret

    def ActivatePane(self, window):
        try:
            ret = aui.AuiManager.ActivatePane(self, window)
        except:
            ret = None
        return ret



import urllib
import wx
import time

from .gridtable_base import SchGridTableBase
from pytigon_lib.schhtml.htmlviewer import tdata_from_html
from pytigon_gui.guilib.tools import colour_to_html


class KeyForRec:
    def __init__(self, rec, tabsort):
        self.rec = rec
        self.tabsort = tabsort

    def __lt__(self, other):
        if self.rec[0].data.strip() == "-":
            return 1
        if other.rec[0].data.strip() == "-":
            return -1

        for pos in self.tabsort:
            if pos > 0:
                x = self.rec[pos - 1].data.lower() < other.rec[pos - 1].data.lower()
            else:
                x = (
                    self.rec[-1 * pos - 1].data.lower()
                    > other.rec[-1 * pos - 1].data.lower()
                )
            if x:
                return x
        return False


def make_key(tabsort):
    return lambda rec: KeyForRec(rec, tabsort)


class PageData(object):
    def __init__(self, parent, page_len, count, titles, first_page):
        self.count = count
        self.page_len = page_len
        self.pages = {}
        self.pages[0] = first_page
        self.parent = parent
        self.titles = titles
        self.inserted = []
        self.sizes = []
        if first_page:
            self.calculate_sizes(titles, 0, first_page)

    def get_page(self, nr):
        href = self.parent.GetParent().get_parm_obj().address
        if "?" in href:
            addr = href + "&page=" + str(nr + 1)
        else:
            addr = href + "?page=" + str(nr + 1)
        html = self.parent.load_data_from_server(addr).decode("utf-8")
        tab = tdata_from_html(html, wx.GetApp().http)
        if tab:
            self.calculate_sizes(tab[0], nr, tab[1:], True)
            return tab[1:]
        else:
            return []

    def __getitem__(self, id):
        if (
            len(self.inserted) > 0
            and self.count > 0
            and id >= self.count - len(self.inserted)
        ):
            return self.inserted[id - (self.count - len(self.inserted))]

        page = id // self.page_len
        id2 = id % self.page_len
        if page in self.pages:
            try:
                ret = self.pages[page][id2]
            except:
                ret = None
            return ret
        else:
            wx.BeginBusyCursor()

            def y(self):
                time.sleep(0.1)

            old_y = wx.GetApp().Yield
            wx.GetApp().Yield = y
            data = self.get_page(page)
            self.pages[page] = data
            wx.GetApp().Yield = old_y
            wx.EndBusyCursor()
            return self.__getitem__(id)

    def sort(self, key):
        data = []
        for i in range(0, self.count):
            data.append(self.__getitem__(i))
        data.sort(key=key)
        self.pages = {}
        self.pages[0] = data
        self.page_len = self.count

    def __len__(self):
        return self.count

    def set_rec(self, id, row):
        if (
            len(self.inserted) > 0
            and self.count > 0
            and id >= self.count - len(self.inserted)
        ):
            self.inserted[id - (self.count - len(self.inserted))] = row
            return

        page = id // self.page_len
        id2 = id % self.page_len
        if page in self.pages:
            try:
                self.pages[page][id2] = row
            except:
                pass

    def append_row(self, row):
        self.inserted.append(row)
        self.count += 1

    def calculate_sizes(self, titles, start_id, page, refresh_if_changed=False):
        l = len(titles)
        row_h = {}
        changed = False
        if not self.sizes:
            self.sizes = [0] * l
        j = start_id - 1
        for row in [
            titles,
        ] + page:
            for i in range(0, l):
                s = len(row[i].data)
                if j == start_id - 1:
                    s *= 1.25
                if s > 64:
                    s = 64
                if s > self.sizes[i]:
                    self.sizes[i] = s
                    changed = True
                if j >= start_id:
                    h = 0
                    if "\n" in row[i].data:
                        h = row[i].data.count("\n")
                    elif len(row[i].data) > 64:
                        h += 1 + len(row[i].data) // 64
                    if h > 0:
                        if j not in row_h or row_h[j] < h:
                            row_h[j] = h
            j += 1
        if refresh_if_changed and changed:
            self.parent.grid.set_col_width(self.sizes)

        if row_h:

            def set_h():
                nonlocal row_h, self
                for key, value in row_h.items():
                    self.parent.grid.SetRowSize(
                        key, value * self.parent.grid.GetDefaultRowSize()
                    )

            wx.CallAfter(set_h)

    def get_sizes(self):
        return self.sizes


class SimpleDataTable(SchGridTableBase):
    def __init__(self, parent, tab):
        SchGridTableBase.__init__(self)
        self._parent = parent
        self.init_data(tab)
        # self.last_row_count = len(self.data)
        self.last_row_count = self.count

    def init_data(self, tab):
        self.colLabels = tab[0]
        self.dataTypes = []
        for col in self.colLabels:
            self.dataTypes.append("s")
        l = tab[0][0].data.split(":")
        self.count = len(tab) - 1
        self.per_page = 0
        if len(l) > 1:
            pages = l[1].split("/")
            if len(pages) > 1:
                self.per_page = int(pages[0])
                self.count = int(pages[1])
                self.colLabels[0].data = l[0]

        if self.per_page > 0:
            self.simple_data = False
            # if self.count > 128:
            #    self.auto_size = "short"
            #    self.data = PageData(
            #        self._parent,
            #        int(self.per_page),
            #        int(self.per_page),
            #        tab[0],
            #        tab[1 : self.per_page + 1],
            #    )
            # else:
            self.data = PageData(
                self._parent,
                int(self.per_page),
                self.count,
                tab[0],
                tab[1 : self.per_page + 1],
            )
        else:
            self.data = PageData(
                self._parent,
                len(tab) - 1,
                len(tab) - 1,
                tab[0],
                tab[1:],
            )
            # self.data = tab[1:]
            # self.per_page = len(self.data)
            self.simple_data = True
        self.attrs = {}
        self.enabled = False

    def replace_tab(self, new_tab):
        SchGridTableBase.replace_tab(self, new_tab)
        self.init_data(new_tab)
        self.refr_count(self.count)

    def refresh_page_data(self, tab):
        old_data = self.data
        self.data = PageData(
            old_data.parent,
            int(self.per_page),
            int(len(tab) - 1),
            tab[0],
            tab[1 : self.per_page + 1],
        )

    def enable(self, enabled):
        self.enabled = enabled
        if not self.simple_data:
            self.data.count = self.count

    def get_action_list(self, row, col=None):
        try:
            if col == None:
                col2 = self.GetNumberCols()
            else:
                col2 = col
            attrs = []
            td = self.data[row][col2]
            attrs.append(td.attrs)
            children = td.children
            if children:
                for sys_id in sorted(list(children)):
                    a = children[sys_id].attrs
                    txt = ""
                    for atom in children[sys_id].atom_list.atom_list:
                        txt += atom.data
                    a["data"] = txt
                    attrs.append(a)
            if col2 != 0:
                attrs2 = self.get_action_list(row, 0)
                for pos in attrs2:
                    attrs.append(pos)
            return attrs
        except:
            return []

    def sort(self, column, append):
        SchGridTableBase.sort(self, column, append)
        if self.count < 4096:
            key = make_key(self.tabsort)
            self.data.sort(key=key)
        else:
            pass
        self.GetView().ForceRefresh()

    def get_ext_attr(self, row, col):
        try:
            tdattr = {}
            td = self.data[row][col]
            tdattr.update(td.attrs)
            if td.children:
                for sys_id in td.children:
                    child = td.children[sys_id]
                    tag = child.tag
                    if tag in tdattr:
                        tdattr[tag] += (child.attrs,)
                    else:
                        tdattr[tag] = (child.attrs,)
        except:
            print("GetExtAttr:", row, col)
            tdattr = None
        return tdattr

    def get_children(self, row, col):
        try:
            td = self.data[row][col]
            return td.children
        except:
            return None

    def filter_cmp(self, pos, key):
        if self.filter_id >= 0:
            if str(pos[self.filter_id].data).upper().startswith(key.upper()):
                return True
            else:
                return False
        else:
            return False

    def GetAttr(self, row, col, kind):
        if row >= self.GetNumberRows():
            attr = self.attr_normal
            attr.IncRef()
            return attr
        try:
            tdattr = self.data[row][col].attrs
        except:
            tdattr = None
            # print("<<<")
            # print("rows:", self.GetNumberRows())
            # print("cols:", self.GetNumberCols())
            # attr = self.attr_normal
            # attr.IncRef()
            # return attr
            # print("len:", len(self.data[row]))
            # print(self.data[row][col])
            # print("Error", row, col)
            # print(">>>")
        if not tdattr:
            attr = self.attr_normal
            attr.IncRef()
            return attr
        bgcolor = None
        color = None
        strong = None
        if "bgcolor" in tdattr:
            bgcolor = tdattr["bgcolor"]
            if bgcolor[0] != "#":
                bgcolor = None
        if "color" in tdattr:
            color = tdattr["color"]
            if color[0] != "#":
                color = None
        if "strong" in tdattr:
            strong = "s"
        if self._is_sel(row):
            bgcolor = colour_to_html(self.sel_colour)
            strong = "s"

        key = ""
        key += bgcolor if bgcolor else "_"
        key += color if color else "_"
        key += strong if strong else "_"
        if "align" in tdattr:
            if tdattr["align"].lower() == "center":
                key += "c"
            elif tdattr["align"].lower() == "right":
                key += "r"
            else:
                key += "_"
        else:
            key += "_"

        if key:
            if key in self.attrs:
                attr = self.attrs[key]
            else:
                attr = wx.grid.GridCellAttr()
                if bgcolor:
                    attr.SetBackgroundColour(bgcolor)
                if color:
                    attr.SetTextColour(color)
                if strong:
                    font = self.GetView().GetDefaultCellFont()
                    font.SetWeight(wx.FONTWEIGHT_BOLD)
                    attr.SetFont(font)
                self.attrs[key] = attr
        else:
            attr = self.attr_normal

        if "align" in tdattr:
            if tdattr["align"].lower() == "center":
                attr.SetAlignment(wx.ALIGN_CENTER, -1)
            elif tdattr["align"].lower() == "right":
                attr.SetAlignment(wx.ALIGN_RIGHT, -1)

        attr.IncRef()
        return attr

    def _get_number_rows(self):
        return self.count

    def GetNumberCols(self):
        if len(self.colLabels) > 0:
            if self.no_actions:
                return len(self.colLabels)
            else:
                return len(self.colLabels) - 1
        else:
            return 0

    def GetValue(self, row, col):
        try:
            ret = self.data[row][col].data
            return ret
        except IndexError:
            return ""

    def SetValue(self, row, col, value):
        try:
            self.data[row][col] = value
        except IndexError:
            self.data.append([""] * self.GetNumberCols())
            self.SetValue(row, col, value)
            msg = wx.grid.GridTableMessage(
                self, wx.grid.GRIDTABLE_NOTIFY_ROWS_APPENDED, 1
            )
            self.GetView().ProcessTableMessage(msg)

    def GetColLabelValue(self, col):
        return self.colLabels[col].data

    def GetTypeName(self, row, col):
        return self.dataTypes[col]

    def CanGetValueAs(self, row, col, type_name):
        col_type = self.dataTypes[col].split(":")[0]
        if type_name == col_type:
            return True
        else:
            return False

    def GetColNames(self):
        return [item.data for item in self.colLabels]

    def CanSetValueAs(self, row, col, type_name):
        return self.CanGetValueAs(row, col, type_name)

    def refresh(self, storePos):
        self.GetView().GetParent().get_parent_form().any_parent_command("refresh_html")

    def set_rec(self, id, row):
        return self.data.set_rec(id, row)

    def append_row(self, row):
        ret = self.data.append_row(row)
        self.count += 1
        return ret

    def get_sizes(self):
        if self.data:
            return self.data.get_sizes()
        else:
            return None

    def copy(self):
        href_base = self._parent.GetParent().get_parm_obj().address
        href = urllib.parse.urljoin(href_base, "../table_action/")
        if self.rec_selected:
            href += "?pk=" + ",".join([str(pos.data) for pos in self.get_sel_rows()[0]])
        data = {
            "action": "copy",
        }
        http = wx.GetApp().get_http(self._parent)
        response = http.post(self._parent, href, parm=data, json_data=True)
        try:
            s = response.json()
        except:
            s = None
        return s

    def paste(self, data):
        href_base = self._parent.GetParent().get_parm_obj().address
        href = urllib.parse.urljoin(href_base, "../table_action/")

        data2 = {"action": "paste", "data": data}
        http = wx.GetApp().get_http(self._parent)
        response = http.post(self._parent, href, parm=data2, json_data=True)
        try:
            s = response.json()
        except:
            s = None
        return s



import os
import wx
import pytigon_gui.guictrl.ctrl

from pytigon_gui.guilib.tools import import_plugin


class SchBaseFrame(wx.Frame):
    """
    This is main window of pytigon application
    """

    def __init__(
        self,
        parent,
        gui_style="tree(toolbar,statusbar)",
        id=-1,
        title="",
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.DEFAULT_FRAME_STYLE | wx.CLIP_CHILDREN | wx.WANTS_CHARS,
        name="MainWindow",
    ):
        wx.Frame.__init__(self, None, wx.ID_ANY, title, pos, size, style, "MainWindow")
        self.run_on_close = []
        self.Bind(wx.EVT_CLOSE, self.on_close)

    def init_plugins(self):
        home_dir = wx.GetApp().get_working_dir()
        app_plugins = os.path.join(wx.GetApp().cwd, "plugins")
        dirnames = [
            os.path.join(wx.GetApp().pytigon_path, "appdata", "plugins"),
            os.path.join(wx.GetApp().data_path, "plugins"),
            app_plugins,
        ]
        auto_plugins = wx.GetApp().config["Global settings"]["auto_plugins"].split(";")
        for dirname in dirnames:
            if not os.path.exists(dirname):
                continue
            for ff in os.listdir(dirname):
                if os.path.isdir(os.path.join(dirname, ff)):
                    dirname2 = os.path.join(dirname, ff)
                    pliki = []
                    for f in os.listdir(dirname2):
                        pliki.append(f)
                    pliki.sort()

                    for f in pliki:
                        if True:
                            if os.path.isdir(os.path.join(dirname2, f)):
                                # p = dirname2.split('/')
                                mod_name = ff + "." + f
                                x = mod_name.replace(".", "/")
                                if (
                                    ff == "auto"
                                    or (
                                        wx.GetApp().plugins and x in wx.GetApp().plugins
                                    )
                                    or x in auto_plugins
                                ):
                                    if ".__" in mod_name:
                                        break
                                    mod = import_plugin(mod_name)
                                    if hasattr(mod, "init_plugin"):
                                        destroy = mod.init_plugin(
                                            wx.GetApp(),
                                            self,
                                            self.desktop,
                                            self._mgr,
                                            self.get_menu_bar(),
                                            self.toolbar_interface,
                                            self.aTable,
                                        )
                                        if destroy != None:
                                            self.destroy_fun_tab.append(destroy)
                        # except:
                        #    import traceback
                        #    print("Error load plugin: ", mod_name)
                        #    print(sys.exc_info()[0])
                        #    print(traceback.print_exc())

    def on_close(self, event):
        for fun in self.run_on_close:
            fun(self)
        event.Skip()



"""Module contains helper classes for grid renderers"""

import textwrap
import wx

from pytigon_gui.guilib.image import SchImage


class ExtStringRenderer(wx.grid.GridCellRenderer):
    def __init__(self):
        """Enhanced version of string renderer"""
        wx.grid.GridCellRenderer.__init__(self)

    def Draw(self, grid, attr, dc, rect, row, col, is_selected):
        rect2 = wx.Rect(rect.x + 1, rect.y + 1, rect.width - 2, rect.height - 2)
        dc.SetBackgroundMode(wx.SOLID)
        if grid.IsEnabled():
            if is_selected:
                dc.SetBrush(wx.Brush(grid.GetSelectionBackground(), wx.SOLID))
            else:
                dc.SetBrush(wx.Brush(attr.GetBackgroundColour(), wx.SOLID))
        else:
            dc.SetBrush(
                wx.Brush(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE), wx.SOLID)
            )
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.DrawRectangle(rect.x, rect.y, rect.width, rect.height)
        (h_align, v_align) = attr.GetAlignment()
        dc.SetBackgroundMode(wx.TRANSPARENT)
        if grid.IsEnabled():
            if is_selected:
                dc.SetTextBackground(grid.GetSelectionBackground())
                dc.SetTextForeground(grid.GetSelectionForeground())
            else:
                dc.SetTextBackground(attr.GetBackgroundColour())
                dc.SetTextForeground(attr.GetTextColour())
        else:
            dc.SetTextBackground(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE))
            dc.SetTextForeground(wx.SystemSettings.GetColour(wx.SYS_COLOUR_GRAYTEXT))
        dc.SetFont(attr.GetFont())

        text = self.get_text(grid, row, col)
        if text.find("#:#") >= 0:
            # dc.DrawLabel(text.split("#:#")[1], rect2, alignment=h_align | v_align)
            dc.DrawLabel(text.split("#:#")[1], rect2, alignment=h_align)
        else:
            # dc.DrawLabel(text, rect2, alignment=h_align | v_align)
            dc.DrawLabel(text, rect2, alignment=h_align)

    def get_text(self, grid, row, col):
        return grid.GetCellValue(row, col)

    def GetBestSize(self, grid, attr, dc, row, col):
        text = grid.GetCellValue(row, col)
        dc.SetFont(attr.GetFont())
        (w, h) = dc.GetTextExtent(text)
        return wx.Size(w, h)

    def Clone(self):
        return ExtStringRenderer()


class MultiLineStringRenderer(ExtStringRenderer):
    """Multiline string renderer"""

    def __init__(self, width):
        self.width = width
        ExtStringRenderer.__init__(self)

    def get_text(self, grid, row, col):
        text = grid.GetCellValue(row, col)
        if len(text) > self.width:
            text2 = textwrap.fill(text, self.width)
        else:
            text2 = text
        return text2

    def GetBestSize(
        self,
        grid,
        attr,
        dc,
        row,
        col,
    ):
        text = grid.GetCellValue(row, col)
        if len(text) > self.width:
            text2 = textwrap.wrap(text, self.width)
            dc.SetFont(attr.GetFont())
            (w, h) = dc.GetTextExtent("x" * self.width)
            return wx.Size(w, h * len(text2))
        else:
            return ExtStringRenderer.GetBestSize(self, grid, attr, dc, row, col)


class IconAndStringRenderer(MultiLineStringRenderer):
    """String renderer extended for rendering icons"""

    def __init__(self):
        MultiLineStringRenderer.__init__(self, 64)
        self.cache = {}

    def get_image_from_cache(self, image):
        if not image in self.cache:
            self.cache[image] = SchImage(image)
        return self.cache[image].bmp

    def get_image(self, grid, row, col):
        children = grid.GetTable().get_children(row, col)
        if children:
            for child_id in children:
                child = children[child_id]
                if child.tag in ("image", "img"):
                    if "src" in child.attrs:
                        return self.get_image_from_cache(child.attrs["src"])
        return None

    def Draw(self, grid, attr, dc, rect, row, col, is_selected):
        dc.DrawRectangle(rect.x, rect.y, rect.width, rect.height)
        image = self.get_image(grid, row, col)
        image2 = None
        if not image:
            text = grid.GetCellValue(row, col)
            if text == "+":
                image2 = wx.ArtProvider.GetBitmap(
                    wx.ART_ADD_BOOKMARK, wx.ART_TOOLBAR, (24, 24)
                )
                if image2:
                    image = image2
        if image:
            rect2 = wx.Rect(
                rect.x + image.GetWidth(),
                rect.y,
                rect.width - image.GetWidth(),
                rect.height,
            )
        else:
            rect2 = rect
        if not image2:
            ExtStringRenderer.Draw(self, grid, attr, dc, rect2, row, col, is_selected)
        if image:
            dc.DrawBitmap(image, rect.x, rect.y, True)

    def GetBestSize(self, grid, attr, dc, row, col):
        image = self.get_image(grid, row, col)
        if image:
            text = grid.GetCellValue(row, col)
            dc.SetFont(attr.GetFont())
            (w, h) = dc.GetTextExtent(text)
            if image.GetHeight() > h:
                h = image.GetHeight()
            w += image.GetWidth()
            return wx.Size(w, h)
        else:
            text = grid.GetCellValue(row, col)
            if text == "+":
                return wx.Size(24, 24)
            else:
                return MultiLineStringRenderer.GetBestSize(
                    self, grid, attr, dc, row, col
                )


class DateTimeRenderer(wx.grid.GridCellRenderer):
    """DateTime renderer"""

    def __init__(self):
        wx.grid.GridCellRenderer.__init__(self)
        self.best_size = None

    def Draw(self, grid, attr, dc, rect, row, col, is_selected):
        rect2 = wx.Rect(rect.x + 2, rect.y + 5, rect.width - 3, rect.height - 6)
        # rect2 = rect
        dc.SetBackgroundMode(wx.SOLID)
        if grid.IsEnabled():
            if is_selected:
                dc.SetBrush(wx.Brush(grid.GetSelectionBackground(), wx.SOLID))
            else:
                dc.SetBrush(wx.Brush(attr.GetBackgroundColour(), wx.SOLID))
        else:
            dc.SetBrush(
                wx.Brush(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE), wx.SOLID)
            )
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.DrawRectangle(rect.x, rect.y, rect.width, rect.height)
        (h_align, v_align) = attr.GetAlignment()
        dc.SetBackgroundMode(wx.TRANSPARENT)
        if grid.IsEnabled():
            if is_selected:
                dc.SetTextBackground(grid.GetSelectionBackground())
                dc.SetTextForeground(grid.GetSelectionForeground())
            else:
                dc.SetTextBackground(attr.GetBackgroundColour())
                dc.SetTextForeground(attr.GetTextColour())
        else:
            dc.SetTextBackground(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE))
            dc.SetTextForeground(wx.SystemSettings.GetColour(wx.SYS_COLOUR_GRAYTEXT))
        dc.SetFont(attr.GetFont())

        d = grid.GetCellValue(row, col)
        text = d[:16]
        dc.DrawLabel(text, rect2, alignment=h_align | v_align)

    def GetBestSize(self, grid, attr, dc, row, col):
        text = grid.GetCellValue(row, col)
        dc.SetFont(attr.GetFont())
        if not self.best_size:
            if text:
                self.best_size = dc.GetTextExtent(text[:16])
            else:
                return wx.Size(0, 0)
        return wx.Size(self.best_size[0], self.best_size[1])

    def Clone(self):
        return ExtStringRenderer()



import wx

try:
    from wx.adv import Wizard, WizardPageSimple
except:
    from wx.wizard import Wizard, WizardPageSimple

import zipfile
import os
import configparser

from pytigon_lib.schtools.install import Ptig
from pytigon_gui.guilib.tools import create_desktop_shortcut

_ = wx.GetTranslation


def make_page_title(wiz_pg, title):
    sizer = wx.BoxSizer(wx.VERTICAL)
    wiz_pg.SetSizer(sizer)
    title = wx.StaticText(wiz_pg, -1, title)
    title.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
    sizer.Add(title, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
    sizer.Add(wx.StaticLine(wiz_pg, -1), 0, wx.EXPAND | wx.ALL, 5)
    return sizer


class TitledPage(WizardPageSimple):
    def __init__(self, parent, lp, title):
        WizardPageSimple.__init__(self, parent)
        self.sizer = make_page_title(self, title)
        self.lp = lp


class InstallWizard(Wizard):
    def __init__(self, file_name):
        Wizard.__init__(self, None, -1, "Install app")
        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.on_wiz_page_changing)
        self.ptig = Ptig(file_name)

        page1 = TitledPage(self, 1, _("Program description"))
        self.page1 = page1
        page2 = TitledPage(self, 2, _("License"))
        page3 = TitledPage(self, 3, _("Run"))

        r_txt = wx.TextCtrl(
            page1, -1, size=wx.Size(600, 300), style=wx.TE_MULTILINE | wx.TE_READONLY
        )
        r_txt.SetValue(self.ptig.get_readme())
        page1.sizer.Add(r_txt)
        self.FitToPage(page1)
        l_txt = wx.TextCtrl(
            page2,
            -1,
            self.ptig.get_license(),
            size=wx.Size(600, 300),
            style=wx.TE_MULTILINE | wx.TE_READONLY,
        )
        page2.sizer.Add(l_txt)
        page2.sizer.AddSpacer(5)
        self.licence = wx.CheckBox(page2, -1, _("Accept"))
        page2.sizer.Add(self.licence)
        page3.sizer.Add(
            wx.StaticText(page3, -1, _("Run the installed program or cancel"))
        )
        WizardPageSimple.Chain(page1, page2)
        WizardPageSimple.Chain(page2, page3)
        self.GetPageAreaSizer().Add(page1)

    def run(self):
        ret = self.RunWizard(self.page1)
        wx.CallAfter(self.Destroy)
        return ret

    def on_wiz_page_changing(self, event):
        if event.GetPage().lp == 2 and event.GetDirection():
            value = self.licence.GetValue()
            if not value:
                wx.MessageBox(
                    _("You must accept the license or cancel the installation!"),
                    _("Installation"),
                )
                event.Veto()
                return
            else:
                if self.install():
                    return
                else:
                    wx.MessageBox(_("Installation error!"), _("Installation"))
                    event.Veto()
                    return
        if event.GetPage().lp == 3 and not event.GetDirection():
            wx.MessageBox(
                _("Program installed, use the uninstaller!"), _("Installation")
            )
            event.Veto()
            return
        event.Skip()

    def install(self):
        self.ptig.extract_ptig()

        ini_file = os.path.join(self.ptig.extract_to, "install.ini")
        created = False
        if os.path.exists(ini_file):
            config = configparser.ConfigParser()
            config.read(ini_file)
            if "INSTALL" in config.sections():
                install = config["INSTALL"]
                title = install.get("Title", self.ptig.prj_name)
                parameters = install.get("Parameters", "")
                create_desktop_shortcut(self.ptig.prj_name, title, parameters)

        if not created:
            create_desktop_shortcut(self.ptig.prj_name, self.ptig.prj_name)

        return True


def install(ptig_path):
    wizard = InstallWizard(ptig_path)
    if wizard.run():
        return True
    else:
        return False
